import asyncio
import aiomcache

from aiocache.utils import get_cache_value_with_fallbacks


class MemcachedBackend:

    DEFAULT_ENDPOINT = "127.0.0.1"
    DEFAULT_PORT = 11211

    def __init__(self, endpoint=None, port=None, loop=None, **kwargs):
        super().__init__(**kwargs)
        self.endpoint = get_cache_value_with_fallbacks(
            endpoint, from_config="endpoint",
            from_fallback=self.DEFAULT_ENDPOINT, cls=self.__class__)
        self.port = get_cache_value_with_fallbacks(
            port, from_config="port",
            from_fallback=self.DEFAULT_PORT, cls=self.__class__)
        self._loop = loop or asyncio.get_event_loop()
        self.client = aiomcache.Client(self.endpoint, self.port, loop=self._loop)
        self.encoding = "utf-8"

    async def _get(self, key):
        """
        Get a value from the cache. Returns default if not found.

        :param key: str
        :returns: obj in key if found else None
        """
        value = await self.client.get(key)
        if value is not None and self.encoding is not None:
            return bytes.decode(value)
        return value

    async def _multi_get(self, keys):
        """
        Get multi values from the cache. For each key not found it returns a None

        :param key: str
        :returns: list of obj for each key found, else if not found
        """
        values = []
        for value in await self.client.multi_get(*keys):
            if value is not None and self.encoding is not None:
                values.append(bytes.decode(value))
            else:
                values.append(value)
        return values

    async def _set(self, key, value, ttl=0):
        """
        Stores the value in the given key.

        :param key: str
        :param value: obj
        :param ttl: int
        :returns: True
        """
        value = str.encode(value) if isinstance(value, str) else value
        return await self.client.set(key, value, exptime=ttl or 0)

    async def _multi_set(self, pairs, ttl=0):
        """
        Stores multiple values in the given keys.

        :param pairs: list of two element iterables. First is key and second is value
        :param ttl: int
        :returns: True
        """
        tasks = []
        for key, value in pairs:
            value = str.encode(value) if isinstance(value, str) else value
            tasks.append(asyncio.ensure_future(self.client.set(key, value, exptime=ttl or 0)))

        await asyncio.gather(*tasks)

        return True

    async def _add(self, key, value, ttl=0):
        """
        Stores the value in the given key. Raises an error if the
        key already exists.

        :param key: str
        :param value: obj
        :param ttl: int
        :returns: True if key is inserted
        :raises: Value error if key already exists
        """
        ret = await self.client.add(key, str.encode(value), exptime=ttl or 0)
        if not ret:
            raise ValueError(
                "Key {} already exists, use .set to update the value".format(key))

        return True

    async def _exists(self, key):
        """
        Check key exists in the cache.

        :param key: str key to check
        :returns: True if key exists otherwise False
        """
        return await self.client.append(key, b'')

    async def _expire(self, key, ttl):
        """
        Expire the given key in ttl seconds. If ttl is 0, remove the expiration

        :param key: str key to expire
        :param ttl: int number of seconds for expiration. If 0, ttl is disabled
        :returns: True if set, False if key is not found
        """
        return await self.client.touch(key, ttl)

    async def _delete(self, key):
        """
        Deletes the given key.

        :param key: Key to be deleted
        :returns: int number of deleted keys
        """
        return 1 if await self.client.delete(key) else 0

    async def _clear(self, namespace=None):
        """
        Deletes the given key.

        :param namespace:
        :returns: True
        """
        if namespace:
            raise ValueError("MemcachedBackend doesnt support flushing by namespace")
        else:
            await self.client.flush_all()
        return True

    async def _raw(self, command, *args, **kwargs):
        """
        Executes a raw command using the underlying client of memcached. It's under
        the developer responsibility to send the needed args and kwargs.

        :param command: str command to execute
        """
        return await getattr(self.client, command)(*args, **kwargs)



from aiocache.cache import SimpleMemoryCache, RedisCache, MemcachedCache
from aiocache.decorators import cached, multi_cached
from aiocache import settings


__all__ = (
    'settings',
    'cached',
    'multi_cached',
    'RedisCache',
    'SimpleMemoryCache',
    'MemcachedCache',
)



import inspect

from aiocache import SimpleMemoryCache
from aiocache.log import logger
from aiocache.cache import BaseCache
from aiocache.utils import class_from_string
from aiocache.serializers import DefaultSerializer
from aiocache.plugins import BasePlugin


DEFAULT_CACHE = SimpleMemoryCache
DEFAULT_CACHE_KWARGS = {}

DEFAULT_SERIALIZER = DefaultSerializer
DEFAULT_SERIALIZER_KWARGS = {}

DEFAULT_PLUGINS = {}


def set_defaults(class_=SimpleMemoryCache, **kwargs):
    """
    Set the default settings for the cache. If within your project you are working with a Redis
    backend, you can use it as::

        aiocache.settings.set_defaults(
            class_="aiocache.RedisCache", endpoint="127.0.0.1", port=6379, namespace="test")

    Once the call is done, all decorators and instances where those params are not specified, the
    default ones will be picked. The class_ param accepts both str and class types.
    """
    if class_:
        if isinstance(class_, str) and issubclass(class_from_string(class_), BaseCache):
            globals()['DEFAULT_CACHE'] = class_from_string(class_)
        elif inspect.isclass(class_) and issubclass(class_, BaseCache):
            globals()['DEFAULT_CACHE'] = class_
        else:
            raise ValueError(
                "DEFAULT_CACHE must be a str or class subclassing aiocache.cache.BaseCache")
    globals()['DEFAULT_CACHE_KWARGS'] = kwargs


def set_default_serializer(class_=DefaultSerializer, **kwargs):
    """
    Set the default settings for the serializer. If within your project you are working with
    json objects, you may want to call it as::

        aiocache.settings.set_default_serializer(
            class_="aiocache.serializers.JsonSerializer")

    Once the call is done, all decorators and instances where serializer params are not specified,
    the default ones will be picked. The class_ param accepts both str and class types.

    If you define your own serializer, you can also set is a default and pass the desired extra
    params through this call.
    """
    if class_:
        if isinstance(class_, str) and issubclass(class_from_string(class_), DefaultSerializer):
            globals()['DEFAULT_SERIALIZER'] = class_from_string(class_)
        elif inspect.isclass(class_) and issubclass(class_, DefaultSerializer):
            globals()['DEFAULT_SERIALIZER'] = class_
        else:
            raise ValueError(
                "DEFAULT_SERIALIZER must be a str or class \
                subclassing aiocache.serializers.DefaultSerializer")
    globals()['DEFAULT_SERIALIZER_KWARGS'] = kwargs


def set_default_plugins(config):
    """
    Set the default settings for the plugins. If within your project you are working with a
    custom plugin, you may want to call it as::

        aiocache.settings.set_default_plugins([
            {
                'class': MyPlugin,
            },
            {
                'class': 'my_module.OtherPlugin',
                'arg': 1
            }
        })

    Once the call is done, all decorators and instances where plugin params are not specified,
    the default ones will be picked. The class param accepts both str and class types.

    If you define your own plugin, you can also set is a default and pass the desired extra
    params through this call.
    """
    new_plugins = {}
    for plugin in config:
        class_ = plugin.pop('class')
        if isinstance(class_, str) and issubclass(class_from_string(class_), BasePlugin):
            new_plugins[class_from_string(class_)] = plugin
        elif inspect.isclass(class_) and issubclass(class_, BasePlugin):
            new_plugins[class_] = plugin
        else:
            logger.warning(
                "%s must be a str or class subclassing aiocache.plugins.BasePlugin" % class_)

    globals()['DEFAULT_PLUGINS'] = new_plugins


def get_defaults():
    return {
        "DEFAULT_CACHE": DEFAULT_CACHE,
        "DEFAULT_CACHE_KWARGS": DEFAULT_CACHE_KWARGS,
        "DEFAULT_SERIALIZER": DEFAULT_SERIALIZER,
        "DEFAULT_SERIALIZER_KWARGS": DEFAULT_SERIALIZER_KWARGS,
        "DEFAULT_PLUGINS": DEFAULT_PLUGINS,
    }


def set_from_dict(config):
    """
    Set the default settings for aiocache from a dict-like structure. The structure is the
    following::

        {
            "CACHE": {
                "class": "aiocache.RedisCache",
                "endpoint": "127.0.0.1",
                "port": 6379
            },
            "SERIALIZER": {
                "class": "aiocache.serializers.DefaultSerializer"
            },
            "PLUGINS": [
                {
                    "class": "aiocache.plugins.BasePlugin"
                }
            ]
        }

    Of course you can set your own classes there. Any extra parameter you put in the dict will be
    used for new instantiations if those are not set explicitly when calling it. The class param
    accepts both str and class types.

    All keys in the config are optional, if they are not passed the previous defaults will be kept.
    """
    if "CACHE" in config:
        class_ = config['CACHE'].pop("class", None)
        set_defaults(class_=class_, **config['CACHE'])

    if "SERIALIZER" in config:
        class_ = config['SERIALIZER'].pop("class", None)
        set_default_serializer(class_=class_, **config['SERIALIZER'])

    if "PLUGINS" in config:
        set_default_plugins(config=config['PLUGINS'])



import os
import time
import functools
import asyncio

import aiocache

from aiocache.log import logger
from aiocache.utils import get_cache_value_with_fallbacks
from aiocache.backends import SimpleMemoryBackend, RedisBackend, MemcachedBackend


class API:

    CMDS = set()

    @classmethod
    def register(cls, fn):
        API.CMDS.add(fn)
        return fn

    @classmethod
    def unregister(cls, fn):
        API.CMDS.discard(fn)

    @classmethod
    def timeout(cls, fn):
        """
        This decorator sets a maximum timeout for a coroutine to execute. The timeout can be both
        set in the ``self.timeout`` attribute or in the ``timeout`` kwarg of the function call.
        I.e if you have a function ``get(self, key)``, if its decorated with this decorator, you
        will be able to call it with ``await get(self, "my_key", timeout=4)``.

        Use either 0 or None to disable the timeout.
        """
        @functools.wraps(fn)
        async def _timeout(self, *args, timeout=None, **kwargs):
            timeout = timeout or self.timeout
            if timeout == 0 or timeout is None:
                return await fn(self, *args, **kwargs)
            return await asyncio.wait_for(fn(self, *args, **kwargs), timeout)

        return _timeout

    @classmethod
    def aiocache_enabled(cls, fake_return=None):
        """
        Use this decorator to be able to fake the return of the function by setting the
        ``AIOCACHE_DISABLE`` environment variable
        """
        def enabled(fn):
            @functools.wraps(fn)
            async def _enabled(*args, **kwargs):
                if os.getenv('AIOCACHE_DISABLE') == "1":
                    return fake_return
                return await fn(*args, **kwargs)

            return _enabled
        return enabled

    @classmethod
    def plugins(cls, fn):
        @functools.wraps(fn)
        async def _plugins(self, *args, **kwargs):
            start = time.time()
            for plugin in self.plugins:
                await getattr(plugin, "pre_{}".format(fn.__name__))(self, *args, **kwargs)

            ret = await fn(self, *args, **kwargs)

            for plugin in self.plugins:
                await getattr(
                    plugin, "post_{}".format(fn.__name__))(
                        self, *args, took=time.time() - start, ret=ret, **kwargs)
            return ret

        return _plugins


class BaseCache:
    """
    Base class that agregates the common logic for the different caches that may exist. Cache
    related available options are:

    :param serializer: obj with :class:`aiocache.serializers.DefaultSerializer` interface.
    :param plugins: list of :class:`aiocache.plugins.BasePlugin` derived classes.
    :param namespace: string to use as default prefix for the key used in all operations of
        the backend.
    :param timeout: int or float in seconds specifying maximum timeout for the operations to last.
        By default its 5. Use 0 or None if you want to disable it.
    """
    DEFAULT_TIMEOUT = 5

    def __init__(self, serializer=None, plugins=None, namespace=None, timeout=None):
        self.timeout = get_cache_value_with_fallbacks(
            timeout, from_config="timeout",
            from_fallback=self.DEFAULT_TIMEOUT, cls=self.__class__)
        self.namespace = get_cache_value_with_fallbacks(
            namespace, from_config="namespace",
            from_fallback=namespace, cls=self.__class__)
        self.encoding = "utf-8"

        self._serializer = None
        self.serializer = serializer or self.get_default_serializer()

        self._plugins = None
        self.plugins = plugins or self.get_default_plugins()

    def get_default_serializer(self):
        return aiocache.settings.DEFAULT_SERIALIZER()

    def get_default_plugins(self):
        return [plugin(**config) for plugin, config in aiocache.settings.DEFAULT_PLUGINS.items()]

    @property
    def serializer(self):
        return self._serializer

    @serializer.setter
    def serializer(self, value):
        self._serializer = value
        self.encoding = getattr(self._serializer, "encoding", 'utf-8')

    @property
    def plugins(self):
        return self._plugins

    @plugins.setter
    def plugins(self, value):
        self._plugins = value

    @API.register
    @API.aiocache_enabled(fake_return=True)
    @API.timeout
    @API.plugins
    async def add(self, key, value, ttl=None, dumps_fn=None, namespace=None):
        """
        Stores the value in the given key with ttl if specified. Raises an error if the
        key already exists.

        :param key: str
        :param value: obj
        :param ttl: int the expiration time in seconds
        :param dumps_fn: callable alternative to use as dumps function
        :param namespace: str alternative namespace to use
        :param timeout: int or float in seconds specifying maximum timeout
            for the operations to last. None by default
        :returns: True if key is inserted
        :raises:
            - ValueError if key already exists
            - :class:`asyncio.TimeoutError` if it lasts more than self.timeout
        """
        start = time.time()
        dumps = dumps_fn or self._serializer.dumps
        ns_key = self._build_key(key, namespace=namespace)

        await self._add(ns_key, dumps(value), ttl)

        logger.debug("ADD %s %s (%.4f)s", ns_key, True, time.time() - start)
        return True

    async def _add(self, key, value, ttl):
        raise NotImplementedError()

    @API.register
    @API.aiocache_enabled()
    @API.timeout
    @API.plugins
    async def get(self, key, default=None, loads_fn=None, namespace=None):
        """
        Get a value from the cache. Returns default if not found.

        :param key: str
        :param default: obj to return when key is not found
        :param loads_fn: callable alternative to use as loads function
        :param namespace: str alternative namespace to use
        :param timeout: int or float in seconds specifying maximum timeout
            for the operations to last. None by default
        :returns: obj loaded
        :raises: :class:`asyncio.TimeoutError` if it lasts more than self.timeout
        """
        start = time.time()
        loads = loads_fn or self._serializer.loads
        ns_key = self._build_key(key, namespace=namespace)

        value = loads(await self._get(ns_key))

        logger.debug("GET %s %s (%.4f)s", ns_key, value is not None, time.time() - start)
        return value or default

    async def _get(self, key):
        raise NotImplementedError()

    @API.register
    @API.aiocache_enabled(fake_return=[])
    @API.timeout
    @API.plugins
    async def multi_get(self, keys, loads_fn=None, namespace=None):
        """
        Get multiple values from the cache, values not found are Nones.

        :param keys: list of str
        :param loads_fn: callable alternative to use as loads function
        :param namespace: str alternative namespace to use
        :param timeout: int or float in seconds specifying maximum timeout
            for the operations to last. None by default
        :returns: list of objs
        :raises: :class:`asyncio.TimeoutError` if it lasts more than self.timeout
        """
        start = time.time()
        loads = loads_fn or self._serializer.loads

        ns_keys = [self._build_key(key, namespace=namespace) for key in keys]
        values = [loads(value) for value in await self._multi_get(ns_keys)]

        logger.debug(
            "MULTI_GET %s %d (%.4f)s",
            ns_keys,
            len([value for value in values if value is not None]),
            time.time() - start)
        return values

    async def _multi_get(self, keys):
        raise NotImplementedError()

    @API.register
    @API.aiocache_enabled(fake_return=True)
    @API.timeout
    @API.plugins
    async def set(self, key, value, ttl=None, dumps_fn=None, namespace=None):
        """
        Stores the value in the given key with ttl if specified

        :param key: str
        :param value: obj
        :param ttl: int the expiration time in seconds
        :param dumps_fn: callable alternative to use as dumps function
        :param namespace: str alternative namespace to use
        :param timeout: int or float in seconds specifying maximum timeout
            for the operations to last. None by default
        :returns: True
        :raises: :class:`asyncio.TimeoutError` if it lasts more than self.timeout
        """
        start = time.time()
        dumps = dumps_fn or self._serializer.dumps
        ns_key = self._build_key(key, namespace=namespace)

        await self._set(ns_key, dumps(value), ttl)

        logger.debug("SET %s %d (%.4f)s", ns_key, True, time.time() - start)
        return True

    async def _set(self, key, value, ttl):
        raise NotImplementedError()

    @API.register
    @API.aiocache_enabled(fake_return=True)
    @API.timeout
    @API.plugins
    async def multi_set(self, pairs, ttl=None, dumps_fn=None, namespace=None):
        """
        Stores multiple values in the given keys.

        :param pairs: list of two element iterables. First is key and second is value
        :param ttl: int the expiration time of the keys in seconds
        :param dumps_fn: callable alternative to use as dumps function
        :param namespace: str alternative namespace to use
        :param timeout: int or float in seconds specifying maximum timeout
            for the operations to last. None by default
        :returns: True
        :raises: :class:`asyncio.TimeoutError` if it lasts more than self.timeout
        """
        start = time.time()
        dumps = dumps_fn or self._serializer.dumps

        tmp_pairs = []
        for key, value in pairs:
            tmp_pairs.append((self._build_key(key, namespace=namespace), dumps(value)))

        await self._multi_set(tmp_pairs, ttl=ttl)

        logger.debug(
            "MULTI_SET %s %d (%.4f)s",
            [key for key, value in tmp_pairs],
            len(pairs),
            time.time() - start)
        return True

    async def _multi_set(self, pairs, ttl):
        raise NotImplementedError()

    @API.register
    @API.aiocache_enabled(fake_return=0)
    @API.timeout
    @API.plugins
    async def delete(self, key, namespace=None):
        """
        Deletes the given key.

        :param key: Key to be deleted
        :param namespace: str alternative namespace to use
        :param timeout: int or float in seconds specifying maximum timeout
            for the operations to last. None by default
        :returns: int number of deleted keys
        :raises: :class:`asyncio.TimeoutError` if it lasts more than self.timeout
        """
        start = time.time()
        ns_key = self._build_key(key, namespace=namespace)
        ret = await self._delete(ns_key)
        logger.debug("DELETE %s %d (%.4f)s", ns_key, ret, time.time() - start)
        return ret

    async def _delete(self, key):
        raise NotImplementedError()

    @API.register
    @API.aiocache_enabled(fake_return=0)
    @API.timeout
    @API.plugins
    async def exists(self, key, namespace=None):
        """
        Check key exists in the cache.

        :param key: str key to check
        :param namespace: str alternative namespace to use
        :param timeout: int or float in seconds specifying maximum timeout
            for the operations to last. None by default
        :returns: True if key exists otherwise False
        :raises: :class:`asyncio.TimeoutError` if it lasts more than self.timeout
        """
        start = time.time()
        ns_key = self._build_key(key, namespace=namespace)
        ret = await self._exists(ns_key)
        logger.debug("EXISTS %s %d (%.4f)s", ns_key, ret, time.time() - start)
        return ret

    async def _exists(self, key):
        raise NotImplementedError()

    @API.register
    @API.aiocache_enabled(fake_return=False)
    @API.timeout
    @API.plugins
    async def expire(self, key, ttl, namespace=None):
        """
        Set the ttl to the given key. By setting it to 0, it will disable it

        :param key: str key to expire
        :param ttl: int number of seconds for expiration. If 0, ttl is disabled
        :param namespace: str alternative namespace to use
        :param timeout: int or float in seconds specifying maximum timeout
            for the operations to last. None by default
        :returns: True if set, False if key is not found
        """
        start = time.time()
        ns_key = self._build_key(key, namespace=namespace)
        ret = await self._expire(ns_key, ttl)
        logger.debug("EXPIRE %s %d (%.4f)s", ns_key, ret, time.time() - start)
        return ret

    async def _expire(self, key, ttl):
        raise NotImplementedError()

    @API.register
    @API.aiocache_enabled(fake_return=True)
    @API.timeout
    @API.plugins
    async def clear(self, namespace=None):
        """
        Clears the cache in the cache namespace. If an alternative namespace is given, it will
        clear those ones instead.

        :param namespace: str alternative namespace to use
        :param timeout: int or float in seconds specifying maximum timeout
            for the operations to last. None by default
        :returns: True
        :raises: :class:`asyncio.TimeoutError` if it lasts more than self.timeout
        """
        start = time.time()
        ret = await self._clear(namespace)
        logger.debug("CLEAR %s %d (%.4f)s", namespace, ret, time.time() - start)
        return ret

    async def _clear(self, namespace):
        raise NotImplementedError()

    @API.register
    @API.aiocache_enabled()
    @API.timeout
    @API.plugins
    async def raw(self, command, *args, **kwargs):
        """
        Send the raw command to the underlying client.

        :param command: str with the command.
        :param timeout: int or float in seconds specifying maximum timeout
            for the operations to last. None by default
        :returns: whatever the underlying client returns
        :raises: :class:`asyncio.TimeoutError` if it lasts more than self.timeout
        """
        start = time.time()
        ret = await self._raw(command, *args, **kwargs)
        logger.debug("%s (%.4f)s", command, time.time() - start)
        return ret

    async def _raw(self, command, *args, **kwargs):
        raise NotImplementedError()

    def _build_key(self, key, namespace=None):
        if namespace is not None:
            return "{}{}".format(namespace, key)
        if self.namespace is not None:
            return "{}{}".format(self.namespace, key)
        return key


class SimpleMemoryCache(SimpleMemoryBackend, BaseCache):
    """
    :class:`aiocache.backends.SimpleMemoryBackend` cache implementation with
    the following components as defaults:
      - serializer: :class:`aiocache.serializers.DefaultSerializer`
      - plugins: None

    Config options are:

    :param serializer: obj with :class:`aiocache.serializers.DefaultSerializer` interface.
    :param plugins: list of :class:`aiocache.plugins.BasePlugin` derived classes.
    :param namespace: string to use as default prefix for the key used in all operations of
        the backend.
    :param timeout: int or float in seconds specifying maximum timeout for the operations to last.
        By default its 5.
    """
    def __init__(self, **kwargs):
        super().__init__(**kwargs)


class RedisCache(RedisBackend, BaseCache):
    """
    :class:`aiocache.backends.RedisBackend` cache implementation with the
    following components as defaults:
      - serializer: :class:`aiocache.serializers.DefaultSerializer`
      - plugins: None

    Config options are:

    :param serializer: obj with :class:`aiocache.serializers.DefaultSerializer` interface.
    :param plugins: list of :class:`aiocache.plugins.BasePlugin` derived classes.
    :param namespace: string to use as default prefix for the key used in all operations of
        the backend.
    :param timeout: int or float in seconds specifying maximum timeout for the operations to last.
        By default its 5.
    :param endpoint: str with the endpoint to connect to
    :param port: int with the port to connect to
    :param db: int indicating database to use
    :param password: str indicating password to use
    """
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def _build_key(self, key, namespace=None):
        if namespace is not None:
            return "{}{}{}".format(namespace, ":" if namespace else "", key)
        if self.namespace is not None:
            return "{}{}{}".format(self.namespace, ":" if self.namespace else "", key)
        return key

    def __repr__(self):  # pragma: no cover
        return "RedisCache ({}:{})".format(self.endpoint, self.port)


class MemcachedCache(MemcachedBackend, BaseCache):
    """
    :class:`aiocache.backends.MemcachedCache` cache implementation with the following
    components as defaults:
      - serializer: :class:`aiocache.serializers.DefaultSerializer`
      - plugins: None

    Config options are:

    :param serializer: obj with :class:`aiocache.serializers.DefaultSerializer` interface.
    :param plugins: list of :class:`aiocache.plugins.BasePlugin` derived classes.
    :param namespace: string to use as default prefix for the key used in all operations of
        the backend.
    :param timeout: int or float in seconds specifying maximum timeout for the operations to last.
        By default its 5.
    :param endpoint: str with the endpoint to connect to
    :param port: int with the port to connect to
    """
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def _build_key(self, key, namespace=None):
        ns_key = super()._build_key(key, namespace=namespace)
        return str.encode(ns_key)

    def __repr__(self):  # pragma: no cover
        return "MemcachedCache ({}:{})".format(self.endpoint, self.port)



import inspect

import aiocache


def class_from_string(class_path):
    class_name = class_path.split('.')[-1]
    module_name = class_path.rstrip(class_name).rstrip(".")
    return getattr(__import__(module_name, fromlist=[class_name]), class_name)


def get_args_dict(func, args, kwargs):
    defaults = {
        arg_name: arg.default for arg_name, arg in inspect.signature(func).parameters.items()
        if arg.default is not inspect._empty
    }
    args_names = func.__code__.co_varnames[:func.__code__.co_argcount]
    return {**defaults, **dict(zip(args_names, args)), **kwargs}


def get_cache(cache=None, serializer=None, plugins=None, **kwargs):
    cache = cache or aiocache.settings.DEFAULT_CACHE
    serializer = serializer
    plugins = plugins

    instance = cache(
        serializer=serializer,
        plugins=plugins,
        **{**aiocache.settings.DEFAULT_CACHE_KWARGS, **kwargs})
    return instance


def get_cache_value_with_fallbacks(value, from_config, from_fallback, cls=None):
    """
    Given a value, decide wether to use the original one, the config or the default one.
    Priority applies as follows:
        - original if its not None
        - from_config if its not None and the cls coincides with the default cache
        - default value
    :param value: original value.
    :param from_config: str name of the field to retrieve from DEFAULT_CACHE_KWARGS.
    :param from_fallback: fallback value to apply in case none of the previous applied.
    :param cls: class of the caller to match with DEFAULT_CACHE.
    """
    v = from_fallback
    if value is not None:
        v = value
    elif cls and issubclass(aiocache.settings.DEFAULT_CACHE, cls):
        v = aiocache.settings.DEFAULT_CACHE_KWARGS.get(from_config, v)

    return v



import asyncio
import itertools
import aioredis

from aiocache.utils import get_cache_value_with_fallbacks


class RedisBackend:

    pools = {}
    DEFAULT_ENDPOINT = "127.0.0.1"
    DEFAULT_PORT = 6379
    DEFAULT_DB = 0
    DEFAULT_PASSWORD = None
    DEFAULT_POOL_MIN_SIZE = 1
    DEFAULT_POOL_MAX_SIZE = 10

    def __init__(
            self, endpoint=None, port=None, db=None,
            password=None, loop=None, pool_min_size=None, pool_max_size=None, **kwargs):
        super().__init__(**kwargs)
        self.endpoint = get_cache_value_with_fallbacks(
            endpoint, from_config="endpoint",
            from_fallback=self.DEFAULT_ENDPOINT, cls=self.__class__)
        self.port = get_cache_value_with_fallbacks(
            port, from_config="port",
            from_fallback=self.DEFAULT_PORT, cls=self.__class__)
        self.db = get_cache_value_with_fallbacks(
            db, from_config="db",
            from_fallback=self.DEFAULT_DB, cls=self.__class__)
        self.password = get_cache_value_with_fallbacks(
            password, from_config="password",
            from_fallback=self.DEFAULT_PASSWORD, cls=self.__class__)
        self.pool_min_size = get_cache_value_with_fallbacks(
            pool_min_size, from_config="pool_min_size",
            from_fallback=self.DEFAULT_POOL_MIN_SIZE, cls=self.__class__)
        self.pool_max_size = get_cache_value_with_fallbacks(
            pool_max_size, from_config="pool_max_size",
            from_fallback=self.DEFAULT_POOL_MAX_SIZE, cls=self.__class__)
        self._loop = loop or asyncio.get_event_loop()

    async def _get(self, key):
        """
        Get a value from the cache

        :param key: str
        :returns: obj in key if found else None
        """

        with await self._connect() as redis:
            return await redis.get(key)

    async def _multi_get(self, keys):
        """
        Get multi values from the cache. For each key not found it returns a None

        :param key: str
        :returns: list of obj for each key found, else if not found
        """
        with await self._connect() as redis:
            return await redis.mget(*keys)

    async def _set(self, key, value, ttl=None):
        """
        Stores the value in the given key.

        :param key: str
        :param value: obj
        :param ttl: int
        :returns: True
        """
        with await self._connect() as redis:
            return await redis.set(key, value, expire=ttl)

    async def _multi_set(self, pairs, ttl=None):
        """
        Stores multiple values in the given keys.

        :param pairs: list of two element iterables. First is key and second is value
        :param ttl: int
        :returns: True
        """
        ttl = ttl or 0

        with await self._connect() as redis:
            transaction = redis.multi_exec()
            flattened = list(itertools.chain.from_iterable(
                (key, value) for key, value in pairs))
            transaction.mset(*flattened)
            if ttl > 0:
                for key in flattened[::2]:
                    transaction.expire(key, timeout=ttl)

            await transaction.execute()

        return True

    async def _add(self, key, value, ttl=None):
        """
        Stores the value in the given key. Raises an error if the
        key already exists.

        :param key: str
        :param value: obj
        :param ttl: int
        :returns: True if key is inserted
        :raises: Value error if key already exists
        """

        with await self._connect() as redis:
            was_set = await redis.set(key, value, expire=ttl, exist=redis.SET_IF_NOT_EXIST)
            if not was_set:
                raise ValueError(
                    "Key {} already exists, use .set to update the value".format(key))
            return was_set

    async def _exists(self, key):
        """
        Check key exists in the cache.

        :param key: str key to check
        :returns: True if key exists otherwise False
        """
        with await self._connect() as redis:
            exists = await redis.exists(key)
            return True if exists > 0 else False

    async def _expire(self, key, ttl):
        """
        Expire the given key in ttl seconds. If ttl is 0, remove the expiration

        :param key: str key to expire
        :param ttl: int number of seconds for expiration. If 0, ttl is disabled
        :returns: True if set, False if key is not found
        """
        with await self._connect() as redis:
            if ttl == 0:
                return await redis.persist(key)
            return await redis.expire(key, ttl)

    async def _delete(self, key):
        """
        Deletes the given key.

        :param key: Key to be deleted
        :returns: int number of deleted keys
        """
        with await self._connect() as redis:
            return await redis.delete(key)

    async def _clear(self, namespace=None):
        """
        Deletes the given key.

        :param namespace:
        :returns: True
        """
        with await self._connect() as redis:
            if namespace:
                keys = await redis.keys("{}:*".format(namespace))
                await redis.delete(*keys)
            else:
                await redis.flushdb()
        return True

    async def _raw(self, command, *args, **kwargs):
        """
        Executes a raw command using the underlying client of aioredis. It's under
        the developer responsibility to send the needed args and kwargs.

        :param command: str command to execute
        """
        with await self._connect() as redis:
            return await getattr(redis, command)(*args, **kwargs)

    def get_pool(self):
        pool_key = "{}{}{}{}{}{}".format(
            self.endpoint, self.port, getattr(self, "encoding", None),
            self.db, self.password, id(self._loop))
        return pool_key, RedisBackend.pools.get(pool_key)

    async def _connect(self):
        pool_key, pool = self.get_pool()

        if pool is None:
            pool = await aioredis.create_pool(
                (self.endpoint, self.port),
                db=self.db,
                password=self.password,
                loop=self._loop,
                encoding=getattr(self, "encoding", None),
                minsize=self.pool_min_size,
                maxsize=self.pool_max_size)
            RedisBackend.pools[pool_key] = pool

        return await pool



from setuptools import setup, find_packages

setup(
    name="aiocache",
    version="0.3.3",
    author="Manuel Miranda",
    url="https://github.com/argaen/aiocache",
    author_email="manu.mirandad@gmail.com",
    description="multi backend asyncio cache",
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
    ],
    packages=find_packages(),
    install_requires=[
        'aioredis',
        'aiomcache'
    ]
)
